/**
 * @(#)Visualiser.java  12/12/99
 * @author Peter Au
 * All rights reserved.
 */

package vishnu.testvis.treemap.visual;

import vishnu.testvis.object.*;
import vishnu.testvis.treemap.rectangle.*;
import java.lang.Exception;
import java.awt.*;
import java.awt.image.ImageObserver;
import java.awt.geom.GeneralPath;

public class Visualiser extends Canvas implements Runnable
{

  static private final int __FRAMEDELAY = 100;

  static final int __FRAME=10;
  static float __borderwidth=5F, __borderheight=5F;
//-- the following variables specify the actual drawing area
  final private float _upperX, _upperY, _width, _height;

  private DataManager _data;
  private Image _treemapView = null;  //-- static treemap view
  private Image[] _zoomingImages;

//-- rectangles represent the cluster-groups and the clusters
  private ClusterGroupRect[] _gpRects = null;
  private ClusterRect[][] _cRects = null;

//-- the rectangle the mouse pointed to, used for tipbox
  private ClusterRect _previousRect = null;

//-- to identify which rectangle has a corner-paint
  private ClusterRect _cornerRect = null;


  /**
   * Constructor: visualise the data into the specific size of pane
   * @param d is the data obj
   * @param x is the left upper x-coordinate of the display pane
   * @param y is the left upper y-coordinate of the display pane
   * @param w is the width of the display pane
   * @param h is the height of the display pane
   */
  public Visualiser(DataManager d, float x, float y, float w, float h)
  {
    int noOfGroup;
    _data = d;
    _upperX=x; _upperY=y; _width=w; _height=h;

    //-- allocate memory for _cRects and _gpRects
    noOfGroup = _data.getClusterGroupsSize();
    _gpRects = new ClusterGroupRect[noOfGroup];
    _cRects = new ClusterRect[noOfGroup][];
    for (int i=0; i<noOfGroup; i++)
      _cRects[i] = new ClusterRect [_data.getClusterGroup(i).getClustersSize()];
    // create rectangles for the corresponding groups and clusters
    createRects();
  }


  static public void setBorder(float w, float h)
  {
    __borderwidth = w;
    __borderheight = h;
  }


  public DataManager getDataManager(){return _data;}

  //---------------------------------------------------------------------
  /**
   * set different variables back to default values
   */
  public void setDefault()
  {
    _previousRect = null;
    _cornerRect = null;
    try
    {
      for (int i=0;i<_gpRects.length; i++)
        if (_gpRects[i].isChosen())
          for(int j=0;j<_cRects[i].length;j++)
            if (_cRects[i][j].isChosen())
	    {
              _gpRects[i].resetChosen();
              _cRects[i][j].resetChosen();
              throw new Exception();
            }  //-- if
    }
    catch(Exception e)
    {
    }  //-- catch
  }


  //---------------------------------------------------------------------
  /**
   * @return a snapshot of the tree map
   */
  public Image getTreemapView(){
    return _treemapView;
  }


  /**
   * create a treemap view of the obj, only call once
   * @param img
   */
  public void createTreemapView(Image img)
  {
  	final float GP_BORDERWIDTH = 1F; // <-- set the width of the border of each group
    _treemapView = img;
    Graphics2D g = (Graphics2D) _treemapView.getGraphics();
    for (int i=0;i<_gpRects.length; i++)
	{
      for(int j=0;j<_cRects[i].length;j++)
	  {
        _cRects[i][j].paint(g);
      }
      _gpRects[i].paint(g);
    } //-- for
    //-- paint the border of each group
    for (int i=0;i<_gpRects.length; i++)
       _gpRects[i].paintBorder(g,GP_BORDERWIDTH);
    // paint the outer border
    ZRect fooRect = new ZRect(_upperX,_upperY,_upperX+_width,_upperY+_height,null);
    g.setColor(Color.black);
    fooRect.paintBorder(g, GP_BORDERWIDTH);

  }  //-- create_pict



  //---------------------------------------------------------------------
  /**
   * draw the treemap view on the image, together with the corner triangle
   */
  public void drawTreemapView(Image img, ImageObserver observer)
  {
    Graphics2D g = (Graphics2D) img.getGraphics();
    g.drawImage(_treemapView,0,0,observer);
    drawCorner(g);
  }


  /**
   * draw a small triangle on the top-right corner of the _cornerRect and make
   * sure that the tipbox is drawn as well
   */
  public void drawCorner(ClusterRect r, Image offScreenImg, ImageObserver observer,
    int x, int y, float w, float h)
  {
    _cornerRect = r;
    _previousRect = null; //-- force the tipbox to be repainted
    drawTipBox(offScreenImg, observer, x, y, w, h);
  }


  /**
   * draw a small triangle on the top-right corner of the _cornerRect
   */
  private void drawCorner(Graphics2D g)
  {
    if (_cornerRect!=null)
    {
      GeneralPath p;
      float x=_cornerRect.getLowerRightX(), y=_cornerRect.getUpperLeftY(), w=8F;
      p = new GeneralPath(GeneralPath.WIND_NON_ZERO);
      if (Math.min(_cornerRect.getWidth(),_cornerRect.getHeight())<w)
        w = (float) Math.min(_cornerRect.getWidth(),_cornerRect.getHeight());
      p.moveTo(x-0.7F,   y+0.9F);
      p.lineTo(x-0.7F-w, y+0.9F);
      p.lineTo(x-0.7F,   y+0.9F+w);
      p.closePath();
      g.setColor(Color.black);
      g.fill(p);
    }
  }


  /**
   * signal the caller (by returning ture) to update the screen, a tipbox
   * with the label of the cluster is drawn on the offscreen-image accordingly
   * @param offScreenImg is the image where the tip box to be drawn on
   * @param observer
   * @param x is the horizontal x position of the mouse event
   * @param y is the vertical y position  of the mouse event
   * @param w is the max width of the image to be used
   * @param h is the max height of the image to be used
   * @return true if the tipbox is drawn on the offScreenImg, otherwise false
   */
  public boolean drawTipBox(Image offScreenImg, ImageObserver observer,
    int x, int y, float w, float h)
  {
    Graphics2D g = (Graphics2D) offScreenImg.getGraphics();
    ClusterRect tmpRect=null;

    //-- find out which rectangle the mouse points to
    try{
      for (int i=0;i<_gpRects.length; i++)
        if (_gpRects[i].contains(x,y))
          for(int j=0;j<_cRects[i].length;j++)
            if (_cRects[i][j].contains(x,y))
            {
              tmpRect = _cRects[i][j];
              throw new Exception();
            }  //-- if

      //-- the mouse is not pointed to any of the rectangles
      if (_previousRect==null) return false;
      else
      {
        //-- the previous pict has a tip box => draw the orginal pict
        _previousRect = null;
        drawTreemapView(offScreenImg, observer);
        return true;
      }
    }
    catch(Exception e){
      //-- the mouse is pointed to a cluster's rectangle

      if (_previousRect==tmpRect)
      {
        //-- if the tip box is already on the graphic, we don't need to update
        return false;
      }
      else
      {
        _previousRect = tmpRect;
        drawTreemapView(offScreenImg, observer);
        if (_previousRect!=null)
        {
          _previousRect.paintLabels(g, w, h);
        }
        return true;
      }

    }  //-- catch
  }  //-- drawTipBox



  //---------------------------------------------------------------------
  /**
   * find out which rectangle the mouse point to (for inspect)
   * @param x is the x-cordinate of the mouse position
   * @param y is the y-cordinate of the mouse position
   * @return a cluster rectangle that the mouse pointed to
   */
  public ClusterRect inspect(int x, int y)
  {
    return findClusterRect(x, y, false);
  }

  /**
   * find out which rectangle the mouse point to (for drill down)
   * @param x is the x-cordinate of the mouse position
   * @param y is the y-cordinate of the mouse position
   * @return a cluster rectangle that the mouse pointed to
   */
  public ClusterRect drillDown(int x, int y)
  {
    return findClusterRect(x, y, true);
  }

  /**
   * @param x is the horizontal x position of the mouse event
   * @param y is the vertical y position  of the mouse event
   * @param setChosenFlag
   * @return the cluster rectangle that the mouse points to, and if the
   *   'setChosenFlag' is true, then set up the corresponding flag in
   *   _gpRect and _cRect
   */
  private ClusterRect findClusterRect(int x, int y, boolean setChosenFlag)
  {
    //-- find out which rectangle the mouse points to
    for (int i=0;i<_gpRects.length; i++)
        if (_gpRects[i].contains(x,y))
            for(int j=0;j<_cRects[i].length;j++)
                if (_cRects[i][j].contains(x,y))
                {
                    if (setChosenFlag)
                    {
                        _gpRects[i].setChosen();
                        _cRects[i][j].setChosen();
                    }
                    return _cRects[i][j];
                }  //-- if
    return null;
  }


  //---------------------------------------------------------------------
  /**
   * !!! BUG - Color is not fixed !!!
   * initialise _gpRects and _cRects
   * create_sub_rects is called
   */
  public void createRects()
  {
    float area = _width*_height;  //-- the area of the display
    int noOfGroup = _data.getClusterGroupsSize();
    float totalDocs = (float) _data.getDocCount();
    float lx=_upperX, ly=_upperY, rx=_upperX+_width, ry=_upperY+_height, tmp=0F;
    Color baseColor = Color.yellow;
    ClusterGroupObj gpObj;
    int gpSize;

    for (int i=0;i<noOfGroup; i++)
    {
        gpObj = _data.getClusterGroup(i);
        gpSize = gpObj.getGroupSize();

    //??? Color is not fixed !!!
        if (i==0) baseColor=Color.cyan;
        if (i==1) baseColor=Color.green;
        if (i==2) baseColor=Color.blue;
        if (i==3) baseColor=Color.orange;
        if (i==4) baseColor=Color.pink;
        if (i==5) baseColor=Color.red;

      //-- West
        if (rx>lx && ry>ly)
        {
            tmp = rx;
            rx = ((float)gpSize/totalDocs)*area/(ry-ly)+lx;

            create_sub_rects(i, rx, ly, lx, ry,
            gpObj.getClusters(), gpSize, baseColor);

            _gpRects[i] = new ClusterGroupRect(lx,ly,rx,ry, baseColor, _data.getClusterGroup(i));
            lx = tmp;
        }
      //-- North
        else if (lx>rx && ry>ly)
        {
            tmp = ry;
            ry = ((float)gpSize/totalDocs)*area/(lx-rx)+ly;

            create_sub_rects(i, rx, ly, lx, ry,
            gpObj.getClusters(), gpSize, baseColor);

            _gpRects[i] = new ClusterGroupRect(lx,ly,rx,ry, baseColor, _data.getClusterGroup(i));
            ly = tmp;
        }
      //-- East
        else if (lx>rx && ly>ry)
        {
            tmp = rx;
            rx = lx - ((float)gpSize/totalDocs)*area/(ly-ry);

            create_sub_rects(i, rx, ly, lx, ry,
                gpObj.getClusters(), gpSize, baseColor);

            _gpRects[i] = new ClusterGroupRect(lx,ly,rx,ry, baseColor, _data.getClusterGroup(i));
            lx = tmp;
        }
      //-- South
        else if (rx>lx && ly>ry)
        {
            tmp = ry;
            ry = ly - ((float)gpSize/totalDocs)*area/(rx-lx);

            create_sub_rects(i, rx, ly, lx, ry,
            gpObj.getClusters(), gpSize, baseColor);

            _gpRects[i] = new ClusterGroupRect(lx,ly,rx,ry, baseColor, _data.getClusterGroup(i));
            ly = tmp;
        }
    } //-- for
  }  //-- createRects


  /**
   * !!! BUG: Color hasn't fixed !!!
   * create the rectangles for the clusters in the specified group
   * @param gpNo is the index of the Group currently working on
   * @param lx is the initial left x-coordinate (based on Rect coordination)
   * @param ly is the initial left y-coordinate (ditto)
   * @param rx is the initial right x-coordinate (ditto)
   * @param ry is the initial right y-coordinate (ditto)
   * @param cs is the list of clusters in the group
   * @param total is the total number of documents found in the group
   * @param col is the base colour of the group
   */
  private void create_sub_rects(int gpNo, float lx, float ly, float rx, float ry,
                             ClusterObj[] cs, int total, Color col)
  {

    ZRect dummy = new ZRect(lx,ly,rx,ry,null);
    float tmp, area = (float) (dummy.getWidth()*dummy.getHeight());

    for(int i=0;i<cs.length;i++)
    {

     if (rx>lx && ry>ly)
     {
       tmp = rx;
        rx = ((float)cs[i].getDocsSize()/(float)total)*area/(ry-ly)+lx;
        _cRects[gpNo][i] = new ClusterRect(lx,ly,rx,ry,col,cs[i]);
        lx = tmp;
      }
      //-- North
      else if (lx>rx && ry>ly)
      {
        tmp = ry;
        ry = ((float)cs[i].getDocsSize()/(float)total)*area/(lx-rx)+ly;
        _cRects[gpNo][i] = new ClusterRect(lx,ly,rx,ry,col,cs[i]);
        ly = tmp;
      }
      //-- East
      else if (lx>rx && ly>ry)
      {
        tmp = rx;
        rx = lx - ((float)cs[i].getDocsSize()/(float)total)*area/(ly-ry);
        _cRects[gpNo][i] = new ClusterRect(lx,ly,rx,ry,col,cs[i]);
        lx = tmp;
      }
      //-- South
      else if (rx>lx && ly>ry)
      {
        tmp = ry;
        ry = ly - ((float)cs[i].getDocsSize()/(float)total)*area/(rx-lx);
        _cRects[gpNo][i] = new ClusterRect(lx,ly,rx,ry,col,cs[i]);
        ly = tmp;
      }

//??? Color hasn't fixed !!!
      col = darker(col);

    }
  }  //-- create_sub_rects


    /**
     * @see        java.awt.Color#darker
     */
    private Color darker(Color c) {
    double FACTOR = 0.85;
      return new Color(Math.max((int)(c.getRed()  *FACTOR), 0),
       Math.max((int)(c.getGreen()*FACTOR), 0),
       Math.max((int)(c.getBlue() *FACTOR), 0));
    }



  //---------------------------------------------------------------------
  //-- pre: _zoomImages can't be null
  public Image getLastZoomingImage()
  {
    return _zoomingImages[_zoomingImages.length-1];
  }

  public Image getZoomingImage(int i){return _zoomingImages[i];}


  public void show_zoomIn(Graphics oldg, ImageObserver observer)
  {
    Graphics2D g = (Graphics2D) oldg;
    for(int i=0;i<_zoomingImages.length;i++)
    {
      //-- time delay for each frame
      timeDelay();
      g.drawImage(_zoomingImages[i],0,0,observer);
    }
    //-- this is time delay for the last frame and the new treemap view
    timeDelay();
  }


  public void show_zoomOut(Graphics oldg, ImageObserver observer)
  {
    Graphics2D g = (Graphics2D) oldg;
    for(int i=_zoomingImages.length-1;i>=0;i--)
    {
      //-- time delay for each frame
      timeDelay();
      g.drawImage(_zoomingImages[i],0,0,observer);
    }
    //-- this is time delay for the last frame and the new treemap view
    timeDelay();
  }



  private void timeDelay()
  {
    Thread delayThread = new Thread(this);
    delayThread.start();
    try{
      delayThread.join();
    }
    catch (InterruptedException ex)
    {
    }
  }


  public void run()
  {
    long lastTime=0, time=0;
    do
    {
      time = System.currentTimeMillis();
      try{
        Thread.sleep(__FRAMEDELAY);
      }
      catch (InterruptedException ex)
      {
      }
      lastTime = time;
    } while (time - lastTime > __FRAMEDELAY);

  }



  //---------------------------------------------------------------------
  /**
   * draw the animated zooming pictures on the buffer
   * @param a list of images that is used for animation pictures
   * @param observer
   */
  public void prepare_zooming(Image[] buffer, ImageObserver observer)
  {
    _zoomingImages = buffer;
    prepare_zooming_buffer(observer);
    create_zooming_picts();
  }

  /**
   * draw the border on each buffer
   * @param a list of buffer
   */
  private void prepare_zooming_buffer(ImageObserver observer)
  {
    Graphics2D g;
    if (_zoomingImages==null)
    {
      System.err.println("Error - zooming buffer cannot be empty!!!");
    }
    else
    {
      //-- copy the static view to one buffer
      g = (Graphics2D) _zoomingImages[0].getGraphics();
      g.drawImage(_treemapView,0,0,observer);
      //-- clear the main treemap area and leave the border
      g.setColor(Color.white);
      g.fill(new Rectangle.Float(_upperX, _upperY, _width, _height));
      //-- copy that buffer to other buffers
      for(int k=1;k<_zoomingImages.length;k++)
      {
        g = (Graphics2D) _zoomingImages[k].getGraphics();
        g.drawImage(_zoomingImages[0],0,0,observer);
      }
    }
  }


  private void create_zooming_picts()
  {
    int[] edgeSize = {0,0,0,0};
    float w_y, e_y, n_x, s_x, portion, tmp;
    ZRect[] tmpRects;
    Graphics tmpGph;
    /**
     * pre-calculate the edgeSize
     * the edgeSize represents the total number of docs on each border
     */
    for(int i=0;i<_gpRects.length;i++)
      if (!_gpRects[i].isChosen())
        edgeSize[_gpRects[i].getDirection()] += _gpRects[i].getCapacity();
      else
        for(int j=0;j<_cRects[i].length;j++)
          if(!_cRects[i][j].isChosen())
            edgeSize[_gpRects[i].getDirection()] += _cRects[i][j].getCapacity();

    //-- initise the starting points on each side
    w_y=_upperY;                e_y=_upperY+__borderheight;
    n_x=_upperX+__borderwidth;  s_x=_upperX;


    for(int i=0;i<_gpRects.length;i++)
    {
      //-- prepare zooming the whole group, if the user has NOT chosen a cluster
      //-- within the group
      if (!_gpRects[i].isChosen())
      {
        portion = ((float)_gpRects[i].getCapacity()) /
                  ((float) edgeSize[_gpRects[i].getDirection()]);
        if (_gpRects[i].isWest())
        {
          tmp = w_y + (portion * (float)(_height-__borderheight));
          _gpRects[i].translateTo(_upperX, w_y, _upperX+__borderwidth, tmp);
          w_y = tmp;
        }
        else if (_gpRects[i].isEast())
        {
          tmp = e_y + (portion * (float)(_height-__borderheight));
          _gpRects[i].translateTo(_upperX+_width-__borderwidth, e_y, _upperX+_width, tmp);
          e_y = tmp;
        }
        else if (_gpRects[i].isNorth())
        {
          tmp = n_x + (portion * (float)(_width-__borderwidth));
          _gpRects[i].translateTo(n_x, _upperY, tmp, _upperY+__borderheight);
          n_x = tmp;

        }
        else
        {
          tmp = s_x + (portion * (float)(_width-__borderwidth));
          _gpRects[i].translateTo(s_x, _upperY+_height-__borderheight, tmp, _upperY+_height);
          s_x = tmp;
        }

        //-- generate a list of animation picture
        tmpRects = _gpRects[i].zooming(_zoomingImages.length);

        //-- draw the rectangles on the offsreen buffer
        for(int j=0;j<_zoomingImages.length;j++)
          tmpRects[j].paint(_zoomingImages[j].getGraphics());

      }
      else
        //-- prepare zooming each cluster within the group
        for (int j=0;j<_cRects[i].length;j++)
          if (!_cRects[i][j].isChosen())
          {
            portion = ((float) _cRects[i][j].getCapacity()) /
                      ((float) edgeSize[_gpRects[i].getDirection()]);
            if (_gpRects[i].isWest())
            {
              tmp = w_y + (portion * (float)(_height-__borderheight));
              _cRects[i][j].translateTo(_upperX, w_y, _upperX+__borderwidth, tmp);
              w_y = tmp;
            }
            else if (_gpRects[i].isEast())
            {
              tmp = e_y + (portion * (float)(_height-__borderheight));
              _cRects[i][j].translateTo(_upperX+_width-__borderwidth, e_y, _upperX+_width, tmp);
              e_y = tmp;
            }
            else if (_gpRects[i].isNorth())
            {
              tmp = n_x + (portion * (float)(_width-__borderwidth));
              _cRects[i][j].translateTo(n_x, _upperY, tmp, _upperY+__borderheight);
              n_x = tmp;
            }
            else
            {
              tmp = s_x + (portion * (float)(_width-__borderwidth));
              _cRects[i][j].translateTo(s_x, _upperY+_height-__borderheight, tmp, _upperY+_height);
              s_x = tmp;
            }  //-- if

            //-- generate a list of animation picture
            tmpRects = _cRects[i][j].zooming(_zoomingImages.length);

            //-- draw the rectangles on the offsreen buffer
            for(int k=0;k<_zoomingImages.length;k++)
              tmpRects[k].paint(_zoomingImages[k].getGraphics());
          }  //-- if
    }  //-- for
  }  //-- prepare_zooming
}
